/********************************************************************************
    LibVideoCoding - video coding library
    Copyright (C) 2002  Dirk Farin

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ********************************************************************************/

#include "libvideocoding/mpeg2/decoder/audio.hh"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include "config.h"


namespace videocoding {

  static inline signed int fixedpoint2int16(mad_fixed_t sample) {
    sample += (1L << (MAD_F_FRACBITS - 16));
    if      (sample >= MAD_F_ONE) sample = MAD_F_ONE - 1;
    else if (sample < -MAD_F_ONE) sample = -MAD_F_ONE;
    return sample >> (MAD_F_FRACBITS + 1 - 16);
  }


  MP2_AudioDecoder::~MP2_AudioDecoder()
  {
    Reset();
    free(d_decoder.sync);
    mad_decoder_finish(&d_decoder);
  }

  void MP2_AudioDecoder::Reset()
  {
    if (d_mad_initialized)
      {
	mad_synth_finish(&d_decoder.sync->synth);
	mad_frame_finish(&d_decoder.sync->frame);
	mad_stream_finish(&d_decoder.sync->stream);
	d_mad_initialized=false;
      }
  }

  bool MP2_AudioDecoder::DoSomething()
  {
    Assert(d_src);
    Assert(d_sink);

    //mad_decoder_run(&d_decoder, MAD_DECODER_MODE_SYNC);
    //return false;

    struct mad_stream *stream;
    struct mad_frame *frame;
    struct mad_synth *synth;

    stream = &d_decoder.sync->stream;
    frame  = &d_decoder.sync->frame;
    synth  = &d_decoder.sync->synth;

    if (!d_mad_initialized)
      {
	mad_stream_init(stream);
	mad_frame_init(frame);
	mad_synth_init(synth);

	mad_stream_options(stream, d_decoder.options);

	d_mad_initialized=true;
      }

    switch (d_decoder.input_func(d_decoder.cb_data, stream)) {
    case MAD_FLOW_STOP:
      return false;
#if 0
    case MAD_FLOW_BREAK:
      goto fail;
    case MAD_FLOW_IGNORE:
      continue;
#endif
    case MAD_FLOW_CONTINUE:
      break;
    }

    while (1) {
      if (d_decoder.header_func) {
	if (mad_header_decode(&frame->header, stream) == -1) {
	  if (!MAD_RECOVERABLE(stream->error))
	    break;

	  (d_decoder.error_func)(d_decoder.cb_data, stream, frame);
	}

	d_decoder.header_func(d_decoder.cb_data, &frame->header);
      }

      if (mad_frame_decode(frame, stream) == -1) {
	if (!MAD_RECOVERABLE(stream->error))
	  break;

	(d_decoder.error_func)(d_decoder.cb_data, stream, frame);
      }

      if (d_decoder.filter_func) {
	d_decoder.filter_func(d_decoder.cb_data, stream, frame);
      }

      mad_synth_frame(synth, frame);

      if (d_decoder.output_func) {
	d_decoder.output_func(d_decoder.cb_data,
			      &frame->header, &synth->pcm);
      }
    }

    return true;
  }


  enum mad_flow MP2_AudioDecoder::input(struct mad_stream* stream)
  {
    // throw away old data

    int used=0;
    if (stream->next_frame != 0)
      used = stream->next_frame-d_buf.AskData();

    if (used)
      {
	uint8* d = d_buf.AskData();
	for (int i=0;i<d_buf.AskLength()-used;i++)
	  d[i] = d[i+used];

	d_buf.TruncateBuffer(used);
      }

    if (d_buf.AskLength()<2000)
      {
      again:
	MP2_SystemPacket* spck = d_src->GetNextPacket();
	if (spck==NULL)
	  return MAD_FLOW_STOP;

	MP2_PES_Packet* pck = dynamic_cast<MP2_PES_Packet*>(spck);
	if (pck==NULL)
	  goto again;

	if (pck)
	  {
	    d_buf.AppendBytes(pck->data.AskData(), pck->data.AskLength());
	    delete pck;

	    mad_stream_buffer(stream, d_buf.AskData(), d_buf.AskLength());
	    return MAD_FLOW_CONTINUE;
	  }
	else
	  {
	    return MAD_FLOW_STOP;
	  }
      }

    return MAD_FLOW_CONTINUE;
  }


  enum mad_flow MP2_AudioDecoder::output(struct mad_header const *header,
					 struct mad_pcm *pcm)
  {
    mad_fixed_t const *left_ch   = pcm->samples[0];
    mad_fixed_t const *right_ch  = pcm->samples[1];

    int16 data[2][2000];
    int idx=0;

    for (int i=0;i<pcm->length;i++)
      {
	data[0][idx] = fixedpoint2int16(*left_ch++);
	data[1][idx] = fixedpoint2int16(*right_ch++);

	idx++;
	if (idx==2000)
	  {
	    d_sink->SendSamples(data[0],data[1],2000);
	    idx=0;
	  }
      }

    if (idx>0) d_sink->SendSamples(data[0],data[1],idx);

    return MAD_FLOW_CONTINUE;
  }


  static enum mad_flow input(void* private_data, struct mad_stream* stream)
  {
    MP2_AudioDecoder* dec = (MP2_AudioDecoder*)(private_data);
    return dec->input(stream);
  }

  static enum mad_flow header(void* private_data, struct mad_header const* header)
  {
    // MP2_AudioDecoder* dec = (MP2_AudioDecoder*)(private_data);

    return MAD_FLOW_CONTINUE;
  }

  static enum mad_flow error(void* private_data,
			     struct mad_stream *stream,
			     struct mad_frame *frame)
  {
    // MP2_AudioDecoder* dec = (MP2_AudioDecoder*)(private_data);

    fprintf(stderr, "decoding error 0x%04x (%s) at byte offset %u\n",
	    stream->error, mad_stream_errorstr(stream),
	    (unsigned int)stream->this_frame); // - ap->shared_mem->netbuf);//((struct buffer *)buffer)->start);

    return MAD_FLOW_IGNORE;
  }

  static enum mad_flow output(void* private_data,
			      struct mad_header const *header,
			      struct mad_pcm *pcm)
  {
    MP2_AudioDecoder* dec = (MP2_AudioDecoder*)(private_data);
    return dec->output(header,pcm);
  }


  MP2_AudioDecoder::MP2_AudioDecoder()
    : d_src(NULL), d_sink(NULL)
  {
    mad_decoder_init(&d_decoder, this, videocoding::input, header, 0, videocoding::output, error, 0);

    ((void*)(d_decoder.sync)) = malloc(sizeof(*d_decoder.sync));

    d_mad_initialized=false;  // (stream specific initialization)
  }

}
