/********************************************************************************
    LibVideoCoding - video coding library
    Copyright (C) 2002  Dirk Farin

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ********************************************************************************/

#include "libvideocoding/mpeg2/decoder/motioncomp.hh"

#include <iostream>

namespace videocoding {

  void Clear(Pixel* yp,int y_stride,int w,int h)
  {
    Pixel c=0;
    if (w==8) c=128;

    for (int y=0;y<h;y++)
      for (int x=0;x<w;x++)
	yp[y*y_stride+x] = c;
  }


  void MP2_MotionCompensation::CompensateAll(const MP2_Macroblock& mb,
					     Bitmap<Pixel>& ref_y,
					     Bitmap<Pixel>& ref_cb,
					     Bitmap<Pixel>& ref_cr,
					     int y0,int x0,
					     MotionVector mv,
					     Pixel* bm_y,  int bm_y_stride,
					     Pixel* bm_cb, int bm_cb_stride,
					     Pixel* bm_cr, int bm_cr_stride,
					     int blkh,
					     bool add)
  {

    // check that motion vector does not point outside of valid reference picture area

    {
      int maxh = ((ref_y.AskWidth()&~15)-x0)<<16;
      int srcx = x0<<16;
      srcx += mv.h;
      if (srcx<0)
	{
	  MessageDisplay::Show(ErrSev_Warning,"motion vector out of valid area (leaves left border)");
	  return;
	}

      if (mv.h>maxh)
	{ 
	  MessageDisplay::Show(ErrSev_Warning,"motion vector out of valid area (leaves right border)");
	  return;
	}

      int maxv = ((ref_y.AskHeight()&~15)-y0)<<16;
      int srcy = y0<<16;
      srcy += mv.v;
      if (srcy<0)
	{
	  MessageDisplay::Show(ErrSev_Warning,"motion vector out of valid area (leaves top border)");
	  return;
	}

      if (mv.v>maxv)
	{ 
	  MessageDisplay::Show(ErrSev_Warning,"motion vector out of valid area (leaves bottom border)");
	  return;
	}
    }


    MotionCompensation* mc_16;
    MotionCompensation* mc_8;

    if (add)
      {
	mc_16 = d_motioncomp_16_avg;
	mc_8  = d_motioncomp_8_avg;
      }
    else
      {
	mc_16 = d_motioncomp_16;
	mc_8  = d_motioncomp_8;
      }

    switch (mb.slicehdr->pichdr->seqhdr->chroma_format)
      {
      case MP2_Chroma_420:
	mc_16->Compensate(bm_y, bm_y_stride,  ref_y, x0,y0, mv,blkh);
	if (mv.h>=0) mv.h >>= 1; else mv.h = (mv.h>>1)+0x4000;
	if (mv.v>=0) mv.v >>= 1; else mv.v = (mv.v>>1)+0x4000;
	mv.h &= ~0x7FFF;  mv.v &= ~0x7FFF;   // currently, this is redundant
	//mv.h /= 2; mv.v /= 2;
	mc_8->Compensate(bm_cb, bm_cb_stride, ref_cb, x0/2,y0/2, mv,blkh/2);
	mc_8->Compensate(bm_cr, bm_cr_stride, ref_cr, x0/2,y0/2, mv,blkh/2);
	break;

#if 0
      case MP2_Chroma_422:
	Compensate16(ref_yp,y0  ,x0  ,mv,yp,y_stride,blkh,add);
	// mv.h /= 2;   
	if (mv.h>=0) mv.h >>= 1; else mv.h = (mv.h>>1)+0x4000;
	mv.h &= ~0x7FFF;
	Compensate8(ref_up,y0  ,x0/2,mv,up,u_stride,blkh,add);
	Compensate8(ref_vp,y0  ,x0/2,mv,vp,v_stride,blkh,add);
	break;

      case MP2_Chroma_444:
	Compensate16(ref_yp,y0  ,x0  ,mv,yp,y_stride,blkh,add);
	Compensate16(ref_up,y0  ,x0  ,mv,up,u_stride,blkh,add);
	Compensate16(ref_vp,y0  ,x0  ,mv,vp,v_stride,blkh,add);
	break;
#endif
      }
  }


  MP2_MotionCompensation::MP2_MotionCompensation(bool accurate)
    //  : d_img_last(NULL), d_img_next(NULL)
  {
    d_motioncomp_16     = GetMotionCompensation(16,false,HalfPelDist, accurate);
    d_motioncomp_8      = GetMotionCompensation( 8,false,HalfPelDist, accurate);
    d_motioncomp_16_avg = GetMotionCompensation(16,true ,HalfPelDist, accurate);
    d_motioncomp_8_avg  = GetMotionCompensation( 8,true ,HalfPelDist, accurate);
  }


  MP2_MotionCompensation::~MP2_MotionCompensation()
  {
    delete d_motioncomp_16;
    delete d_motioncomp_8;
    delete d_motioncomp_16_avg;
    delete d_motioncomp_8_avg;
  }


  const int refbm_fwd = 0;
  const int refbm_bkw = 1;

  const int str_frm = 0;
  const int str_top = 1;
  const int str_bot = 2;


  void MP2_MotionCompensation::InitializeReferences(const Image<Pixel>* last,
						    const Image<Pixel>* next,
						    const Image<Pixel>* curr,
						    bool  P_picture,
						    bool  field_structure,
						    bool  first_field,
						    bool  predicting_top_field)
  {
    Assert(last);
    Assert(next);

    for (int d=0;d<2;d++)
      for (int f=0;f<=2;f++)
	{
	  d_ref_ybm [d][f].Release();
	  d_ref_cbbm[d][f].Release();
	  d_ref_crbm[d][f].Release();
	}


    if (field_structure && !first_field && P_picture)
      {
	for (int f=1;f<=2;f++)
	  {
	    const Image<Pixel>* img;

	    if ((predicting_top_field && f==str_top) || (!predicting_top_field && f==str_bot))
	      img = last;
	    else
	      img = curr;

	    d_ref_ybm [refbm_fwd][f]=img->AskBitmapY ().CreateFieldView(f==str_top);
	    d_ref_cbbm[refbm_fwd][f]=img->AskBitmapCb().CreateFieldView(f==str_top);
	    d_ref_crbm[refbm_fwd][f]=img->AskBitmapCr().CreateFieldView(f==str_top);
	  }
      }
    else
      {
	for (int d=0;d<2;d++)
	  {
	    const Image<Pixel>* img = (d==refbm_fwd ? last : next);

	    d_ref_ybm [d][0]=img->AskBitmapY();
	    d_ref_cbbm[d][0]=img->AskBitmapCb();
	    d_ref_crbm[d][0]=img->AskBitmapCr();

	    d_ref_ybm [d][1]=img->AskBitmapY().CreateFieldView(true);
	    d_ref_cbbm[d][1]=img->AskBitmapCb().CreateFieldView(true);
	    d_ref_crbm[d][1]=img->AskBitmapCr().CreateFieldView(true);
    
	    d_ref_ybm [d][2]=img->AskBitmapY().CreateFieldView(false);
	    d_ref_cbbm[d][2]=img->AskBitmapCb().CreateFieldView(false);
	    d_ref_crbm[d][2]=img->AskBitmapCr().CreateFieldView(false);
	  }
      }
  }


  void MP2_MotionCompensation::CreatePrediction(const MP2_Macroblock& mb,
						int x0,int y0,
						Pixel* yp,int y_stride,
						Pixel* up,int u_stride,
						Pixel* vp,int v_stride)
  {
    bool add=false;

    // iterate for both directions d==0 -> forward / d==1 -> backward
    for (int d=0;d<2;d++)
      {
	//const Image<Pixel>* refimg;

	int pred_direction;

	if (d==0)
	  {
	    // forward

	    if (!mb.motion.has_mv_forward) continue;
	    pred_direction = refbm_fwd; //refimg = d_img_last;
	  }
	else
	  {
	    // backward

	    if (!mb.motion.has_mv_backward) continue;
	    pred_direction = refbm_bkw; //refimg = d_img_next;
	  }

	MotionVector mv[2];
	mv[0] = mb.motion.vector[0][d];
	mv[1] = mb.motion.vector[1][d];

	Bitmap<Pixel> ref_yp[2], ref_up[2], ref_vp[2];

	if (mb.motion.prediction_mode == MP2_Frame_FrameBased)
	  {
	    ref_yp[0] = d_ref_ybm [pred_direction][0];
	    ref_up[0] = d_ref_cbbm[pred_direction][0];
	    ref_vp[0] = d_ref_crbm[pred_direction][0];
	  }
	else
	  {
	    int topfield[2];

	    topfield[0] = ((mb.motion.vertical_field_select[0][d]==MP2_Pred_TopField) ? str_top : str_bot);
	    topfield[1] = ((mb.motion.vertical_field_select[1][d]==MP2_Pred_TopField) ? str_top : str_bot);

	    ref_yp[0] = d_ref_ybm [pred_direction][topfield[0]];
	    ref_up[0] = d_ref_cbbm[pred_direction][topfield[0]];
	    ref_vp[0] = d_ref_crbm[pred_direction][topfield[0]];
					       
	    ref_yp[1] = d_ref_ybm [pred_direction][topfield[1]];
	    ref_up[1] = d_ref_cbbm[pred_direction][topfield[1]];
	    ref_vp[1] = d_ref_crbm[pred_direction][topfield[1]];
	  }

	switch (mb.motion.prediction_mode)
	  {
	  case MP2_Frame_FrameBased:
	    CompensateAll(mb,
			  ref_yp[0],   ref_up[0],   ref_vp[0],    y0,x0, mv[0],
			  yp,y_stride, up,u_stride, vp,v_stride,  16, add);
	    break;

	  case MP2_Frame_FieldBased:
	    CompensateAll(mb,
			  ref_yp[0],     ref_up[0],     ref_vp[0],      y0/2,x0, mv[0],
			  yp         ,y_stride*2, up         ,u_stride*2, vp         ,v_stride*2, 8, add);
	    CompensateAll(mb,
			  ref_yp[1],     ref_up[1],     ref_vp[1],      y0/2,x0, mv[1],
			  yp+y_stride,y_stride*2, up+u_stride,u_stride*2, vp+v_stride,v_stride*2, 8, add);
	    break;

	  case MP2_Field_FieldBased:
	    CompensateAll(mb,
			  ref_yp[0],   ref_up[0],   ref_vp[0],    y0,x0, mv[0],
			  yp,y_stride, up,u_stride, vp,v_stride,  16, add);
	    break;

	  case MP2_Field_16x8MC:
	    CompensateAll(mb,
			  ref_yp[0],     ref_up[0],     ref_vp[0],      y0,x0, mv[0],
			  yp,y_stride,   up,u_stride,   vp,v_stride,    8, add);
	    CompensateAll(mb,
			  ref_yp[1],     ref_up[1],     ref_vp[1],      y0+8,x0, mv[1],
			  yp+8*y_stride,y_stride, up+4*u_stride,u_stride, vp+4*v_stride,v_stride,  8, add);
	    break;

	  case MP2_Frame_DualPrime:
	  case MP2_Field_DualPrime:
	    std::cerr << "Dual Prime\n";
	    Clear(yp,y_stride,16,16);
	    Clear(up,u_stride,8,8);
	    Clear(vp,v_stride,8,8);
	    assert(0); // TODO
	    exit(10);
	    break;
	  }

	add=true;
      }
  }

}
