/********************************************************************************
    LibVideoCoding - video coding library
    Copyright (C) 2002  Dirk Farin

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ********************************************************************************/

#include "pessource.hh"
#include "libvideocoding/mpeg2/syntax/errors.hh"
#include <libvideogfx/utility/bitstream/bitreader.hh>


namespace videocoding {

  void MP2_ProgramStream_Decoder::SetInputStream(InputStream& istr)
  {
    d_istr = &istr;
  }


  MP2_SystemPacket* MP2_ProgramStream_Decoder::GetNextPacket()
  {
    Assert(d_istr);
    uint8 buf[4];

  again:
    uint64 pos = d_istr->AskPosition();   // remember position where packet started
    MP2_SystemPacket* pck = NULL;         // pointer to the new system packet

    SkipZerosToNextStartCode(*d_istr);     // skip leading zeros before start code

    int size = d_istr->FillBuffer(buf,4,4);  // read start code (or invalid junk)

    if (size==0) return NULL;   // end of stream

    try
      {
	if (size<4)
	  throw Excpt_UnexpectedEndOfData("program stream start-code");

	if (buf[0] != 0x00 || buf[1] != 0x00 || buf[2] != 0x01)
	  throw Excpt_InvalidData("start code expected but junk data found");

	switch (buf[3])
	  {
	  case MP2_StartCode_ISO11172_End:
	    pck = new MP2_Iso11172EndCode;
	    break;

	  case MP2_StartCode_PackStart:
	    d_istr->Pushback(buf,4);
	    {
	      MP2_PackHeader* pack = new MP2_PackHeader;
	      pck = pack;
	      pack->Read(*d_istr);
	    }
	    break;

	  case MP2_StartCode_SystemHeaderStart:
	    d_istr->Pushback(buf,4);
	    {
	      MP2_SystemHeader* syshdr = new MP2_SystemHeader;
	      pck = syshdr;
	      syshdr->Read(*d_istr);
	    }
	    break;

	  default:
	    if (buf[3] >= 0xBC)
	      {
		d_istr->Pushback(buf,4);
		{
		  MP2_PES_Packet* pespck = new MP2_PES_Packet;
		  pck = pespck;
		  pespck->Read(*d_istr);
		}
	      }
	    else
	      throw Excpt_InvalidData("non system-header start code found");

	    break;
	  }
      }
    catch (const Excpt_BitstreamError& e)
      {
	// show bitstream error, destroy packet and try again

	if (pck) delete pck;

	MessageDisplay::Show(e);

	SkipToNextStartCode(*d_istr);  // resynchronize
	goto again;
      }
    catch (...)
      {
	if (pck) delete pck;
	pck = NULL;

	throw;
      }

    if (pck)
      pck->input_filepos = pos;

    return pck;
  }


  bool MP2_ProgramStream_Decoder::IsEOF() const
  {
    Assert(d_istr);

    return d_istr->IsEOF();
  }


  // -----------------------------------------------------------


  const int elem_stream_packet_size = 10000;
  const int min_elem_packet_size = 100;


  MP2_ElementaryStream_Decoder::MP2_ElementaryStream_Decoder()
    : d_minimal_len(min_elem_packet_size),
      d_istr(NULL),
      d_streamid(MP2_StartCode_Video0)
  {
  }

  void MP2_ElementaryStream_Decoder::SetInputStream(InputStream& istr)
  {
    d_istr = &istr;
  }


  MP2_SystemPacket* MP2_ElementaryStream_Decoder::GetNextPacket()
  {
    Assert(d_istr);

    if (d_istr->IsEOF())
      return NULL;

    MP2_PES_Packet* pespck = new MP2_PES_Packet;
    pespck->Reset();
    pespck->stream_id = d_streamid;
    pespck->input_filepos = d_istr->AskPosition();
    pespck->data_start_offset = 0;

    uint8* buf = pespck->data.GetPtrToAppendToBuffer(elem_stream_packet_size);
    int len = d_istr->FillBuffer(buf,elem_stream_packet_size,d_minimal_len);

    Assert(len>0);  // if there is no more input, IsEOF() above should have returned true

    pespck->data.TruncateBuffer(elem_stream_packet_size-len);

    return pespck;
  }


  bool MP2_ElementaryStream_Decoder::IsEOF() const
  {
    Assert(d_istr);

    return d_istr->IsEOF();
  }


  // -----------------------------------------------------------

  MP2_UniversalStream_Decoder::MP2_UniversalStream_Decoder()
    : d_src(NULL),
      d_type(MPEG_Unknown_Stream),
      d_forced_type(false)
  {
  }

  void MP2_UniversalStream_Decoder::ForceStreamType(MP2_Stream_Type type)
  {
    d_type = type;
    d_forced_type = true;

    if (d_type==MPEG_Unknown_Stream) d_forced_type=false;
  }

  MP2_Stream_Type MP2_UniversalStream_Decoder::AskStreamType()
  {
    DetermineType();

    return d_type;
  }

  void MP2_UniversalStream_Decoder::SetInputStream(InputStream& str)
  {
    d_src = &str;

    d_elem.SetInputStream(str);
    d_prog.SetInputStream(str);

    if (!d_forced_type) d_type = MPEG_Unknown_Stream;
  }

  MP2_SystemPacket* MP2_UniversalStream_Decoder::GetNextPacket()
  {
    DetermineType();

    switch (d_type)
      {
      case MPEG_Audio_Stream:
      case MPEG_Video_Stream: return d_elem.GetNextPacket(); break;
      case MPEG_Program_Stream: return d_prog.GetNextPacket(); break;
      case MPEG_Transport_Stream: Assert(0); break; // TODO
      case MPEG_Unknown_Stream: return NULL; break;

      default:
	Assert(0);
	break;
      }
  }

  bool  MP2_UniversalStream_Decoder::IsEOF() const
  {
    Assert(d_src);

    const_cast<MP2_UniversalStream_Decoder*>(this)->DetermineType();

    switch (d_type)
      {
      case MPEG_Audio_Stream:
      case MPEG_Video_Stream: return d_elem.IsEOF(); break;
      case MPEG_Program_Stream: return d_prog.IsEOF(); break;
      case MPEG_Transport_Stream: Assert(0); break; // TODO
      case MPEG_Unknown_Stream: return true; break;

      default:
	Assert(0);
	break;
      }
  }


  void MP2_UniversalStream_Decoder::DetermineType()
  {
    Assert(d_src);

    if (d_type==MPEG_Unknown_Stream)
      {
	d_type = DetermineStreamType(*d_src);

	if (d_type == MPEG_Video_Stream) d_elem.SetStreamID(MP2_StartCode_Video0);
	if (d_type == MPEG_Audio_Stream) d_elem.SetStreamID(MP2_StartCode_Audio0);
      }
  }


  // -----------------------------------------------------------

  MP2_PES_JunkFilter::MP2_PES_JunkFilter()
    : d_src(NULL)
  {
  }


  void MP2_PES_JunkFilter::SetSource(MP2_PES_Source& src)
  {
    d_src = &src;
  }


  MP2_SystemPacket* MP2_PES_JunkFilter::GetNextPacket()
  {
    Assert(d_src);

    // read until we found a non-junk packet
    for (;;)
      {
	// read a packet

	MP2_SystemPacket* p = d_src->GetNextPacket();
	if (p==NULL)
	  return NULL;

	// this class only works with PES packets

	MP2_PES_Packet* pes = dynamic_cast<MP2_PES_Packet*>(p);
	Assert(pes);

	// if packet is too short, throw it away

	if (pes->data.AskLength()<4)
	  {
	    delete pes;
	    continue;
	  }

	// if it is a valid packet that begins with a start code, return it

	uint8* data = pes->data.AskData();
	if (data[0]==0 && data[1]==0 && data[2]==1)
	  return pes;

	// else throw it away

	delete pes;
      }
  }


  bool MP2_PES_JunkFilter::IsEOF() const
  {
    Assert(d_src);
    return d_src->IsEOF();
  }

}
